from __future__ import print_function
# coding: utf-8
# Copyright (c) Max-Planck-Institut für Eisenforschung GmbH - Computational Materials Design (CM) Department
# Distributed under the terms of "New BSD License", see the LICENSE file.

import sys
from pyiron_base import GenericJob
from pyiron_contrib.protocol.utils import IODictionary, InputDictionary, LoggerMixin, Event, EventHandler, \
    Pointer, CrumbType, ordered_dict_get_last, Comparer, TimelineDict
# from pyiron_contrib.protocol.utils.types import PyironJobTypeRegistry
from pyiron_contrib.protocol.utils.pptree import print_tree as pptree
from abc import ABC, abstractmethod


"""
The objective is to iterate over a directed acyclic graph of simulation instructions.
"""

__author__ = "Liam Huber, Dominik Gehringer, Jan Janssen"
__copyright__ = "Copyright 2019, Max-Planck-Institut für Eisenforschung GmbH " \
                "- Computational Materials Design (CM) Department"
__version__ = "0.0"
__maintainer__ = "Liam Huber"
__email__ = "huber@mpie.de"
__status__ = "development"
__date__ = "Aug 16, 2019"


# defines the name, for which a subclass of Protocol will be searched in order to get a default whitelist configuration
DEFAULT_WHITE_LIST_ATTRIBUTE_NAME = 'DefaultWhitelist'


class Vertex(LoggerMixin, ABC):
    """
    A parent class for objects which are valid vertices of a directed acyclic graph.

    Attributes:
        input (InputDictionary): A pointer-capable dictionary for inputs, including a sub-dictionary for defaults.
            (Default is a clean dictionary.)
        output (IODictionary): A pointer-capable dictionary for outputs. (Default is a clean dictionary.)
        archive (IODictionary): A pointer-capable dictionary for sampling the history of inputs and outputs. (Default
            is a clean dictionary.)
        vertex_state (str): Which edge to follow out of this vertex. (Default is "next".)
        possible_vertex_states (list[str]): Allowable exiting edge names. (Default is ["next"], one edge only!)
        vertex_name (str): Vertex name. (Default is None, gets set automatically when if the vertex is being added to
            a graph, or if the vertex is a Protocol being instantiated.)
        n_history (int): The length of each list stored in the output dictionary. (Default is 1, keep only the most
            recent output.)
        on (bool): Whether to execute the vertex when it is the active vertex of the graph, or simply skip over it.
            (default is True -- actually execute!)
        graph_parent (Vertex): The object who owns the graph that this vertex resides in. (Default is None.)

    Input attributes:
        default (IODictionary): A dictionary for fall-back values in case a key is requested that isn't in the main
            input dictionary.

    Archive attributes:
        whitelist (IODictionary): A nested dictionary of periods for archiving input and output values. Stores on
            executions where `clock % period = 0`.
        clock (int): The timer for whether whether or not input/output should be archived to hdf5.
    """

    def __init__(self, **kwargs):
        try:  # Super magic when the inheritance path is just vertices
            super(Vertex, self).__init__()
        except TypeError:  # Super magic when the inheritance path includes GenericJob (i.e. for a Protocol)
            super(Vertex, self).__init__(**kwargs)
        self.input = InputDictionary()
        self.output = IODictionary()
        self.archive = IODictionary()
        self.archive.clock = 0
        self.archive.output = IODictionary()
        self.archive.input = IODictionary()
        self.archive.whitelist = IODictionary()
        self.archive.whitelist.input = IODictionary()
        self.archive.whitelist.output = IODictionary()
        self._vertex_state = "next"
        self.possible_vertex_states = ["next"]
        self.vertex_name = None
        self.n_history = 1
        self.on = True
        self.graph_parent = None

    def get_graph_location(self):
        return self._get_graph_location()[:-1]  # Cut the trailing underscore

    def _get_graph_location(self, loc=""):
        new_loc = self.vertex_name + "_" + loc
        if self.graph_parent is None:
            return new_loc
        else:
            return self.graph_parent._get_graph_location(loc=new_loc)

    @property
    def vertex_state(self):
        return self._vertex_state

    @vertex_state.setter
    def vertex_state(self, new_state):
        if new_state not in self.possible_vertex_states:
            raise ValueError("New state not in list of possible states")
        self._vertex_state = new_state

    @abstractmethod
    def execute(self):
        """What to do when this vertex is the active vertex during graph traversal."""
        pass

    @property
    def whitelist(self):
        return {
            'input': self.archive.whitelist.input,
            'output': self.archive.whitelist.output
        }

    @whitelist.setter
    def whitelist(self, value):
        self.set_whitelist(value)

    def set_whitelist(self, dictionary):
        """
        Sets whitelist of the current vertex. Argument defines the form:
        ```
            {'input': 5,
             'output': 1}
            # sets all keys of input to dump at every fith execution cycle
            # sets all keys of output to dump at every execution cycle
            {'input': {'structure': None,
                        'forces': 5}
            }
            # disables the archiveing of input.structure but keeps forces
        ```
        Args:
            dictionary (dict): The whitelist specification.
        """
        for k, v in dictionary.items():
            if k not in ('input', 'output'):
                raise ValueError

            if isinstance(v, int):
                self._set_archive_period(k, v)
            elif isinstance(v, dict):
                self._set_archive_whitelist(k, **v)
            else:
                raise TypeError

    def _set_archive_whitelist(self, archive, **kwargs):
        """
        Whitelist properties of either "input" or "output" archive and set their dump period.

        Args:
            archive (str): either 'input' or 'output'.
            **kwargs: property names, values should be positive integers, specifies the dump freq, None = inf = < 0.

        """
        for k, v in kwargs.items():
            whitelist = getattr(self.archive.whitelist, archive)
            whitelist[k] = v

    def _set_archive_period(self, archive, n, keys=None):
        """
        Sets the archive period for each property of to "n" if keys is not specified.
        If keys is a list of property names, "n" will be set a s archiving period only for those

        Args:
            archive (str): Either 'input' or 'output'.
            n (int): Dump at every `n` steps
            keys (list of str): The affected keys

        """
        if keys is None:
            keys = list(getattr(self, archive).keys())
        self._set_archive_whitelist(archive, **{
            k: n for k in keys
        })

    def set_input_archive_period(self, n, keys=None):
        self._set_archive_period('input', n, keys=keys)

    def set_output_archive_period(self, n, keys=None):
        self._set_archive_period('output', n, keys=keys)

    def set_input_whitelist(self, **kwargs):
        self._set_archive_whitelist('input', **kwargs)

    def set_output_whitelist(self, **kwargs):
        self._set_archive_whitelist('output', **kwargs)

    def set_archive_period(self, n):
        self.set_input_archive_period(n)
        self.set_output_archive_period(n)

    def _update_archive(self):
        # Update input
        history_key = 't_%s' % self.archive.clock

        for key, value in self.input.items():
            if key in self.archive.whitelist.input:
                # the keys there, but it could be explicitly set to < 0 or None
                period = self.archive.whitelist.input[key]
                if period is not None and period >= 0:
                    # TODO: Notifaction when whitelist contains items which are not items of input
                    # check if the period matches that of the key
                    if self.archive.clock % period == 0:
                        if key not in self.archive.input:
                            self.archive.input[key] = TimelineDict()
                            self.archive.input[key][history_key] = value
                        else:
                            # we want to archive it only if there is a change, thus get the last element
                            last_val = ordered_dict_get_last(self.archive.input[key])
                            if not Comparer(last_val) == value:
                                self.archive.input[key][history_key] = value
                                self.logger.info('Property "{}" did change in input ({} -> {})'.format(
                                    key, last_val, value
                                ))
                            else:
                                self.logger.info('Property "{}" did not change in input'.format(key))

        # Update output
        for key, value in self.output.items():
            if key in self.archive.whitelist.output:
                period = self.archive.whitelist.output[key]
                if period is not None and period >= 0:
                    # TODO: Notifaction when whitelist contains items which are not items of input
                    # check if the period matches that of the key
                    if self.archive.clock % period == 0:
                        val = value[-1]
                        if key not in self.archive.output:
                            self.archive.output[key] = TimelineDict()
                            self.archive.output[key][history_key] = val
                        else:
                            # we want to archive it only if there is a change, thus get the last element
                            last_val = ordered_dict_get_last(self.archive.output[key])
                            if not Comparer(last_val) == val:
                                self.archive.output[key][history_key] = val
                            else:
                                self.logger.info('Property "{}" did not change in input'.format(key))

    def _update_output(self, output_data):
        if output_data is None:
            return

        for key, value in output_data.items():
            if key not in self.output:
                self.output[key] = [value]
            else:
                history = list(self.output[key])
                # Roll the list if it is necessary
                history.append(value)
                if len(history) > self.n_history:
                    # Remove the head of the queue
                    history.pop(0)
                self.output[key] = history

    def update_and_archive(self, output_data):
        self._update_output(output_data)
        self._update_archive()

    def finish(self):
        pass

    def parallel_setup(self):
        """How to prepare to execute in parallel when there's a list of these vertices together."""
        pass

    def to_hdf(self, hdf, group_name=None):
        """
        Store the Vertex in an HDF5 file.

        Args:
            hdf (ProjectHDFio): HDF5 group object.
            group_name (str): HDF5 subgroup name. (Default is None.)
        """
        if group_name is not None:
            hdf5_server = hdf.open(group_name)
        else:
            hdf5_server = hdf

        hdf5_server["TYPE"] = str(type(self))
        hdf5_server["possiblevertexstates"] = self.possible_vertex_states
        hdf5_server["vertexstate"] = self.vertex_state
        hdf5_server["vertexname"] = self.vertex_name
        hdf5_server["nhistory"] = self.n_history
        self.input.to_hdf(hdf=hdf5_server, group_name="input")
        self.output.to_hdf(hdf=hdf5_server, group_name="output")
        self.archive.to_hdf(hdf=hdf5_server, group_name="archive")

    def from_hdf(self, hdf, group_name=None):
        """
        Load the Vertex from an HDF5 file.

        Args:
            hdf (ProjectHDFio): HDF5 group object.
            group_name (str): HDF5 subgroup name. (Default is None.)
        """
        if group_name is not None:
            hdf5_server = hdf.open(group_name)
        else:
            hdf5_server = hdf

        self.possible_vertex_states = hdf5_server["possiblevertexstates"]
        self._vertex_state = hdf5_server["vertexstate"]
        self.vertex_name = hdf5_server["vertexname"]
        self.n_history = hdf5_server["nhistory"]
        self.input.from_hdf(hdf=hdf5_server, group_name="input")
        self.output.from_hdf(hdf=hdf5_server, group_name="output")
        self.archive.from_hdf(hdf=hdf5_server, group_name="archive")

        # sort the dictionaries after loading, do it for both input and output dictionaries
        for archive_name in ('input', 'output'):
            archive = getattr(self.archive, archive_name)
            for key in archive.keys():
                history = archive[key]
                # create an ordered dictionary from it, convert it to integer back again
                archive[key] = TimelineDict(
                    sorted(history.items(),
                           key=lambda item: int(item[0].replace('t_', '')))
                )


class PrimitiveVertex(Vertex):
    """
    Vertices which do not contain their a sub-graph but directly produce output from input.
    """

    def execute(self):
        """Just parse the input and do your physics, then store the output."""
        output_data = self.command(**self.input.resolve())
        self.update_and_archive(output_data)

    @abstractmethod
    def command(self, *args, **kwargs):
        """The command method controls the physics"""
        pass

    def execute_parallel(self, queue, n, input_dict):
        """How to execute in parallel when there's a list of these vertices together."""
        output_data = self.command(**input_dict)
        queue.put((n, output_data))
        # Note: The output needs to be explicitly collected and archived later if this is used in place of `execute`


class CompoundVertex(Vertex): #, PyironJobTypeRegistry):
    """
    Vertices which contain a graph and produce output only after traversing their graph to its exit point.

    Input:
        graph (Graph): The graph of vertices to traverse.
        protocol_finished (Event):
        protocol_started (Event):
        vertex_processing (Event):
        vertex_processed (Event):
        finished (bool):
    """
    def __init__(self, **kwargs):
        super(CompoundVertex, self).__init__(**kwargs)

        self.graph = Graph()
        self.graph.owner = self

        # Initialize event system
        self.protocol_finished = Event()
        self.protocol_started = Event()
        self.vertex_processing = Event()
        self.vertex_processed = Event()

        # Set up the graph
        self.define_vertices()
        self.define_execution_flow()
        self.define_information_flow()

        # On initialization, set the active verex to starting vertex
        self.graph.active_vertex = self.graph.starting_vertex

        self.finished = False

        self.restore_default_whitelist()

    @property
    def default_whitelist(self):
        # set the default whitelist -> check if the class has an attribute, otherwise skip the configuration
        cls = type(self)
        if hasattr(cls, DEFAULT_WHITE_LIST_ATTRIBUTE_NAME):
            return getattr(cls, DEFAULT_WHITE_LIST_ATTRIBUTE_NAME)
        else:
            return None

    @abstractmethod
    def define_vertices(self):
        """Add child vertices to the graph."""
        pass

    @abstractmethod
    def define_execution_flow(self):
        """Wire the logic for traversing the graph edges."""
        pass

    @abstractmethod
    def define_information_flow(self):
        """Connect input and output information inside the graph. Also set the archive clock for all vertices."""
        pass

    @abstractmethod
    def get_output(self):
        """
        Define the output dictionary to be returned when the graph traversal completes. This synchronizes the
        behaviour of primitive vertices and compound vertices when they themselves are the child vertex in another
        graph.
        """
        pass

    def execute(self):
        """Traverse graph until the active vertex is None."""
        # Subscribe graph vertices to the protocol_finished Event
        for vertex_name, vertex in self.graph.vertices.items():
            handler_name = '{}_close_handler'.format(vertex_name)
            if not self.protocol_finished.has_handler(handler_name):
                self.protocol_finished += EventHandler(handler_name, vertex.finish)

        # Run the graph
        if self.graph.active_vertex is None:
            self.graph.active_vertex = self.graph.starting_vertex
        self.protocol_started.fire()
        while self.graph.active_vertex is not None:
            vertex_on = self.graph.active_vertex.on
            if isinstance(vertex_on, Pointer):
                vertex_on = ~vertex_on
            if not vertex_on:
                self.logger.info('Skipping vertex "{}":{}'.format(self.graph.active_vertex.vertex_name,
                                                                  type(self.graph.active_vertex).__name__))
                self.graph.step()
            self.vertex_processing.fire(self.graph.active_vertex)
            self.graph.active_vertex.execute()
            self.vertex_processed.fire(self.graph.active_vertex)
            self.graph.step()
        self.graph.active_vertex = self.graph.restarting_vertex
        self.update_and_archive(self.get_output())

    def execute_parallel(self, queue, n, input):
        """How to execute in parallel when there's a list of these vertices together."""
        self.execute()
        queue.put((n, self.get_output()))

    def set_graph_archive_clock(self, clock, recursive=False):
        for _, vertex in self.graph.vertices.items():
            vertex.archive.clock = clock
            if recursive and isinstance(vertex, CompoundVertex):
                vertex.set_graph_archive_clock(clock, recursive=True)

    def to_hdf(self, hdf, group_name=None):
        """
        Store the Protocol in an HDF5 file.

        Args:
            hdf (ProjectHDFio): HDF5 group object.
            group_name (str): HDF5 subgroup name - optional
        """
        with hdf.open(self.vertex_name) as graph_hdf:
            self.graph.to_hdf(hdf=graph_hdf, group_name="graph")
        super(CompoundVertex, self).to_hdf(hdf=hdf, group_name=group_name)

    def from_hdf(self, hdf=None, group_name=None):
        """
        Load the Protocol from an HDF5 file.

        Args:
            hdf (ProjectHDFio): HDF5 group object - optional
            group_name (str): HDF5 subgroup name - optional
        """
        super(CompoundVertex, self).from_hdf(hdf=hdf, group_name=group_name)
        self.graph.from_hdf(hdf=hdf, group_name="graph")
        self.define_information_flow()  # Rewire pointers

    def visualize(self, execution=True, dataflow=True):
        return self.graph.visualize(self.fullname(), execution=execution, dataflow=dataflow)

    @property
    def whitelist(self):
        return {
            vertex_name: vertex.whitelist for vertex_name, vertex in self.graph.vertices.items()
        }

    @whitelist.setter
    def whitelist(self, value):
        self.set_whitelist(value)

    def _set_archive_period(self, archive, n, keys=None):
        """
        In constrast to Vertex._set_archive_period, this calls "n". if keys=None it will be applied to all vertices
        Args:
            archive: (str) input or output
            n: (int) dump every "n" steps
            keys: (list of str) the vertex names which the it should be applied to (default: None)
        """
        if keys is None:
            keys = self.graph.vertices.keys()

        for key in keys:
            vertex = self.graph.vertices[key]
            vertex._set_archive_period(archive, n)

    def set_input_archive_period(self, n, keys=None):
        self._set_archive_period('input', n, keys=keys)

    def set_output_archive_period(self, n, keys=None):
        self._set_archive_period('output', n, keys=keys)

    def _set_archive_whitelist(self, archive, **kwargs):
        """
        Sets a whitelist configuration for this protocol
        Args:
            archive: (str) "input" or "output" the archive the whitelist should be applied to
            **kwargs: vertex_name = dict the whitelist configuration
        """
        for key, value in kwargs.items():
            if key not in self.graph.vertices:
                self.logger.warning('Cannot set the whitelist of vertex "%s" since it is no a part of protocol "%s' % (key, self.vertex_name))
                continue
            vertex = self.graph.vertices[key]
            vertex._set_archive_whitelist(archive, **value)

    def set_input_whitelist(self, **kwargs):
        self._set_archive_whitelist('input', **kwargs)

    def set_output_whitelist(self, **kwargs):
        self._set_archive_whitelist('output', **kwargs)

    def set_whitelist(self, dictionary):
        """
        Sets the whitelist for this protocol. The first level of keys must contain valid vertex ids. (A warning will
        be printed otherwise.
            1. Level one: graph vertex ids -> specifies to which vertices the nested whitelist will be applied to
            2. Level two: "input" or "output" -> defines which dictionary is affected by the archiving periods
            3. Level three: the properties of the corresponding dictionary which will be affected
        Example:
        ```
            { 'calc_static' : {
                'input': 5 #sets all keys of input dict to 5,
                'output': {
                    'energy_pot': 1,
                    'structure': None
                }
        ```

        Args:
            dictionary: (dict) the whitelist configuration
        """

        for key, vertex_dict in dictionary.items():
            if key not in self.graph.vertices:
                self.logger.warning(
                    'Cannot set the whitelist of vertex "%s" since it is no a part of protocol "%s' % (key, self.vertex_name))
                continue
            vertex = self.graph.vertices[key]
            # call it for each vertex
            vertex.set_whitelist(vertex_dict)

    def set_archive_period(self, n):
        """
        Sets the archive period for all key for both input and output dictionaries
        Args:
            n: (int) the archiving period
        """
        self.set_input_archive_period(n)
        self.set_output_archive_period(n)

    def restore_default_whitelist(self):
        """
        If the protcol type has an attribute DefaultWhitelist, it will be restored
        """
        if self.default_whitelist is not None:
            # first we have to clear again all white lists
            for vertex_name, vertex in self.graph.vertices.items():
                vertex.archive.whitelist.input.clear()
                vertex.archive.whitelist.output.clear()
            self.whitelist = self.default_whitelist
            self.logger.debug('Whitelist configured for protocol "%s"' % self.vertex_name)

    def format_whitelist(self, format='tree', file=sys.stdout):
        if format == 'code':
            start = [self.vertex_name, 'graph']
            path_format = '{path} = {value}\n'
            for vertex_name, conf in self.whitelist.items():
                vertex_path = start + [vertex_name, 'archive', 'whitelist']
                for archive, vertex_conf in conf.items():
                    archive_path = vertex_path + [archive]
                    for key, value in vertex_conf.items():
                        key_path = archive_path + [key]
                        file.write(path_format.format(path='.'.join(key_path), value=value))
        elif format == 'simple':

            from pyiron_contrib.protocol.utils.pptree import count_paths

            def print_tree(node, level=0):
                if not isinstance(node, dict):
                    return '%s\n' % str(node)
                elif count_paths(node) == 0:
                    return ''
                else:
                    output = ''
                    for key, val in node.items():
                        # only print the path if there is a leaf node
                        if count_paths(val) > 0:
                            output += '%s%s:' % ('\t'*level, key)
                            # now lets do the second part of the comma
                            # add an extra \n if val is a dectionary
                            output += '\n' if isinstance(val, dict) else ' '
                            output += print_tree(val, level=level+1)
                    return output
            # print the result
            file.write(print_tree(self.whitelist))
        elif format == 'tree':
            # print the nice tree
            pptree(self.whitelist, file=file, name='%s.%s' % (self.vertex_name, 'whitelist'))


class Protocol(CompoundVertex, GenericJob):
    """
    A parent class for compound vertices which are being instantiated as regular pyiron jobs, i.e. the highest level
    graph in their context.

    Example: if `X` inherits from `CompoundVertex` and performs the desired logic, then
    ```
    class ProtocolX(Protocol, X):
        pass
    ```
    can be added to the `pyiron_contrib`-level `__init__` file and jobs performing X-logic can be instantiated with
    in a project `pr` with the name `job_name` using `pr.create_job(pr.job_type.ProtocolX, job_name)`.
    """

    def __init__(self, project=None, job_name=None):
        super(Protocol, self).__init__(project=project, job_name=job_name)
        self.vertex_name = job_name

    def execute(self):
        super(Protocol, self).execute()
        self.protocol_finished.fire()

    def run_static(self):
        """If this CompoundVertex is the highest level, it can be run as a regular pyiron job."""
        self.status.running = True
        self.execute()
        self.status.collect = True  # Assume modal for now
        self.run()  # This is an artifact of inheriting from GenericJob, to get all that run functionality

    def run(self, run_again=False, repair=False, debug=False, run_mode=None, continue_run=False):
        """A wrapper for the run which allows us to simply keep going with a new variable `continue_run`"""
        if continue_run:
            self.status.created = True
        super(CompoundVertex, self).run(run_again=run_again, repair=repair, debug=debug, run_mode=run_mode)

    def collect_output(self):
        # Dear Reader: This feels like a hack, but it works. Sincerely, -Liam
        self.to_hdf()

    def write_input(self):
        # Dear Reader: I looked at base/master/list and /parallel where this appears, but it's still not clear to me
        # what I should be using this for. But, I get a NotImplementedError if I leave it out, so here it is. -Liam
        pass

    def to_hdf(self, hdf=None, group_name=None):
        """
        Store the Protocol in an HDF5 file.

        Args:
            hdf (ProjectHDFio): HDF5 group object - optional
            group_name (str): HDF5 subgroup name - optional
        """
        if hdf is None:
            hdf = self.project_hdf5
        self.graph.to_hdf(hdf=hdf, group_name="graph")
        super(CompoundVertex, self).to_hdf(hdf=hdf, group_name=group_name)

    def from_hdf(self, hdf=None, group_name=None):
        """
        Load the Protocol from an HDF5 file.

        Args:
            hdf (ProjectHDFio): HDF5 group object - optional
            group_name (str): HDF5 subgroup name - optional
        """
        if hdf is None:
            hdf = self.project_hdf5
        super(Protocol, self).from_hdf(hdf=hdf, group_name=group_name)


class Graph(dict, LoggerMixin):
    """
    A directed graph of vertices and edges, and a method for iterating through the graph.

    Vertices and edges are the graph are explicitly stored as child classes inheriting from `dict` so that all
    'graphiness' is fully decoupled from the objects sitting at the vertices.

    Attributes:
        vertices (Vertices): Vertices of the graph.
        edges (Edges): Directed edges between the vertices.
        starting_vertex (Vertex): The element of `vertices` for the graph to begin on.
        active_vertex (Vertex): The element of `vertices` for the vertex the graph iteration is on.
        restarting_vertex (Vertex): The element of `vertices` for the graph to restart on if the graph has been loaded.
    """

    def __init__(self, **kwargs):
        super(Graph, self).__init__(**kwargs)
        self.vertices = Vertices()
        self.edges = Edges()
        self.starting_vertex = None
        self.active_vertex = None
        self.restarting_vertex = None
        self.owner = None

    def __setattr__(self, key, val):
        if key == "vertices":
            if not isinstance(val, Vertices):
                raise ValueError("'vertices' is a protected attribute for graphs.")
            self[key] = val
        elif key == "edges":
            if not isinstance(val, Edges):
                raise ValueError("'edges' is a protected attribute for graphs.")
            self[key] = val
        elif key in ["active_vertex", "starting_vertex", "restarting_vertex"]:
            if val is None or isinstance(val, Vertex):
                self[key] = val
            else:
                raise ValueError("The active, starting, and restarting vertices must inherit `Vertex` or be `None`.")
        elif key == "owner":
            if not (isinstance(val, CompoundVertex) or val is None):
                raise ValueError("Only protocols can hold graphs, but the assigned owner has type", type(val))
            else:
                self[key] = val
        elif isinstance(val, Vertex):
            val.vertex_name = key
            val.graph_parent = self.owner
            self.vertices[key] = val
            self.edges.initialize(val)
        else:
            raise TypeError("Graph vertices must inherit from `Vertex`")

    def __getattr__(self, name):
        try:
            return self["vertices"][name]
        except KeyError:
            try:
                return self[name]
            except KeyError:
                return object.__getattribute__(self, name)

    def visualize(self, protocol_name, execution=True, dataflow=True):
        """
        Plot a visual representation of the graph.

        Args:
            protocol_name:
            execution (bool): Show the lines dictating the flow of graph traversal.
            dataflow (bool): Show the lines dictating where vertex input comes from.

        Returns:
            (graphviz.Digraph) The image representation of the protocols workflow
        """
        try:
            from graphviz import Digraph
        except ImportError as import_error:
            self.logger.exception('Failed to import "graphviz" package', exc_info=import_error)
            return

        # Create graph object
        workflow = Digraph(comment=protocol_name)

        # Define styles for the individual classes
        class_style_mapping = {
            CompoundVertex: {'shape': 'box'},
            # CommandBool: {'shape': 'diamond'},
            PrimitiveVertex: {'shape': 'circle'}
        }

        def resolve_type(type_):
            if type_ in class_style_mapping.keys():
                return type_
            else:
                parents = [key for key in class_style_mapping.keys() if issubclass(type_, key)]
                if len(parents) == 0:
                    raise TypeError('I do not know how to visualize "{}"'.format(type_.__name__))
                elif len(parents) > 1:
                    self.logger.warn('More than one parent class found for type "{}"'.format(type_.__name__))
                return parents[0]

        for vertex_name, vertex in self.vertices.items():
            vertex_type = type(vertex)
            vertex_type_style = class_style_mapping[resolve_type(vertex_type)]

            node_label = '''<<B>{vertex_type}</B><BR/>{vertex_name}>'''
            node_label = node_label.format(vertex_type=vertex_type.__name__,
                                           vertex_name=vertex_name)
            if self.active_vertex == vertex:
                # Highlight active vertex
                highlight = {
                    'style': 'filled',
                    'color': 'green',
                }
            else:
                highlight = {}

            # Highlight the active vertex in green color
            highlight.update(vertex_type_style)
            workflow.node(vertex_name, label=node_label, **highlight)
        # Add end node
        workflow.node('end', 'END', **{'shape': 'doublecircle', 'style': 'filled', 'color': 'red'})
        protocol_input_node = None

        if execution:
            for vertex_start, edges in self.edges.items():
                for vertex_state, vertex_end in edges.items():
                    if vertex_end is None:
                        vertex_end = 'end'
                    workflow.edge(vertex_start, vertex_end, label=vertex_state)
        if dataflow:
            dataflow_edge_style = {
                'style': 'dotted',
                'color': 'blue',
                'labelfontcolor': 'blue',
                'labelangle': '90'
            }
            for vertex_name, vertex in self.vertices.items():
                items = super(InputDictionary, vertex.input).items()
                for key, value in items:
                    if isinstance(value, Pointer):
                        vertex_end = self._edge_from_pointer(key, value)
                        if vertex_end is not None:
                            if isinstance(vertex_end, Vertex):
                                workflow.edge(vertex_end.vertex_name, vertex_name, label=key, **dataflow_edge_style)
                            elif isinstance(vertex_end, (IODictionary, Vertex)):
                                self.logger.warning('vertex_end is IODIctionary() I have to decide what to do')
                                if protocol_input_node is None:
                                    # Initialize a node for protocol level input
                                    protocol_input_node = workflow.node(
                                        'protocol_input', '{}.input'.format(protocol_name),
                                        style='filled',
                                        shape='folder'
                                    )
                                workflow.edge('protocol_input', vertex_name, label=key, **dataflow_edge_style)
                            else:
                                pass

        return workflow

    def _edge_from_pointer(self, key, p):
        assert isinstance(p, Pointer)
        path = p.path.copy()
        root = path.pop(0)

        result = root.object
        while len(path) > 0:
            crumb = path.pop(0)
            crumb_type = crumb.crumb_type
            crumb_name = crumb.name

            if isinstance(result, (Vertex, IODictionary)):
                return result

            # If the result is a pointer itself we have to resolve it first
            if isinstance(result, Pointer):
                self.logger.info('Resolved pointer in a pointer path')
                result = ~result
            if isinstance(crumb_name, Pointer):
                self.logger.info('Resolved pointer in a pointer path')
                crumb_name = ~crumb_name
            # Resolve it with the correct method - dig deeper
            if crumb_type == CrumbType.Attribute:
                try:
                    result = getattr(result, crumb_name)
                except AttributeError as e:
                    raise e
            elif crumb_type == CrumbType.Item:
                try:
                    result = result.__getitem__(crumb_name)
                except (TypeError, KeyError) as e:
                    raise e

        # If we reached this point we have no Command at all, give a warning
        self.logger.warning('I could not find a graph in the pointer {} for key "{}"'.format(p.path, key))
        return None

    def step(self):
        """
        Follows the edge out of the active vertex to get the name of the next vertex and set it as the active vertex.
        If the active vertex has multiple possible states, the outbound edge for the current state will be chosen.

        Returns:
            (str) The name of the next vertex.
        """
        vertex = self.active_vertex
        if vertex is not None:
            state = vertex.vertex_state
            next_vertex_name = self.edges[vertex.vertex_name][state]

            if next_vertex_name is None:
                self.active_vertex = None
            else:
                self.active_vertex = self.vertices[next_vertex_name]

    def make_edge(self, start, end, state="next"):
        """
        Makes a directed edge connecting two vertices.

        Args:
            start (Vertex): The vertex for the edge to start at.
            end (Vertex): The vertex for the edge to end at.
            state (str): The state for the vertex to be in when it points to this particular end. (Default, "next", is
                the parent-level state for vertices without multiple outbound edges.)
        """
        if start.vertex_name not in self.vertices.keys():
            raise ValueError("The vertex {} was not found among graph vertices, {}".format(start.vertex_name,
                                                                                           self.vertices.keys()))

        if state not in start.possible_vertex_states:
            raise ValueError("{} not found in possible vertex states for {}, {}".format(
                state,
                start.vertex_name,
                start.possible_vertex_states
            ))

        if end is not None:
            if end.vertex_name not in self.vertices.keys():
                raise ValueError("{} is not among vertices, {}".format(end.vertex_name, self.vertices.keys()))
            self.edges[start.vertex_name][state] = end.vertex_name
        else:
            self.edges[start.vertex_name][state] = None

    def make_pipeline(self, *args):
        """
        Adds an edge between every argument, in the order they're given. The edge is added for the vertex state "next",
        so this is only appropriate for vertices which don't have a non-trivial `vertex_state`.

        Args:
            *args (Vertex/str): Vertices to connect in a row, or the state connecting two vertices.
        """
        for n, vertex in enumerate(args[:-1]):
            if isinstance(vertex, str):
                continue
            next_vertex = args[n + 1]
            if isinstance(next_vertex, str):
                state = next_vertex
                next_vertex = args[n + 2]
            else:
                state = "next"
            self.make_edge(vertex, next_vertex, state=state)

    def to_hdf(self, hdf, group_name="graph"):
        with hdf.open(group_name) as hdf5_server:
            hdf5_server["TYPE"] = str(type(self))
            hdf5_server["startingvertexname"] = self.starting_vertex.vertex_name
            if self.active_vertex is not None:
                hdf5_server["activevertexname"] = self.active_vertex.vertex_name
            else:
                hdf5_server["activevertexname"] = None
            if self.restarting_vertex is None:
                hdf5_server["restartingvertexname"] = self.starting_vertex.vertex_name
            else:
                hdf5_server["restartingvertexname"] = self.restarting_vertex.vertex_name
            self.vertices.to_hdf(hdf=hdf5_server, group_name="vertices")
            self.edges.to_hdf(hdf=hdf5_server, group_name="edges")

    def from_hdf(self, hdf, group_name="graph"):
        with hdf.open(group_name) as hdf5_server:
            active_vertex_name = hdf5_server["activevertexname"]
            self.vertices.from_hdf(hdf=hdf5_server, group_name="vertices")
            self.edges.from_hdf(hdf=hdf5_server, group_name="edges")
            self.starting_vertex = self.vertices[hdf5_server["startingvertexname"]]
            self.restarting_vertex = self.vertices[hdf5_server["restartingvertexname"]]
        if active_vertex_name is not None:
            self.active_vertex = self.vertices[active_vertex_name]
        else:
            self.active_vertex = self.restarting_vertex


class Vertices(dict):
    """
    A dictionary of vertices whose keys are the vertex name.
    """

    def __init__(self, **kwargs):
        super(Vertices, self).__init__(**kwargs)

    def __setattr__(self, key, val):
        if not isinstance(val, Vertex):
            raise ValueError
        self[key] = val

    def __getattr__(self, item):
        return self[item]

    def to_hdf(self, hdf, group_name="vertices"):
        with hdf.open(group_name) as hdf5_server:
            hdf5_server["TYPE"] = str(type(self))
            for name, vertex in self.items():
                if isinstance(vertex, Vertex):
                    vertex.to_hdf(hdf=hdf5_server, group_name=name)
                else:
                    raise TypeError("Cannot save non-Vertex-like vertices")

    def from_hdf(self, hdf, group_name="vertices"):
        with hdf.open(group_name) as hdf5_server:
            for name, vertex in self.items():
                if isinstance(vertex, (CompoundVertex, Vertex)):
                    vertex.from_hdf(hdf=hdf5_server, group_name=name)
                else:
                    raise TypeError("Cannot load non-Vertex-like vertices")


class Edges(dict):
    """
    A collection of dictionaries connecting each state of a given vertex to another vertex.
    """

    def __init__(self, **kwargs):
        super(Edges, self).__init__(**kwargs)

    def __setattr__(self, key, val):
        self[key] = val

    def __getattr__(self, item):
        return self[item]

    def initialize(self, vertex):
        """
        Set an outbound edge to `None` for each allowable vertex state.

        Args:
            vertex (Vertex): The vertex to assign an edge to.
        """
        name = vertex.vertex_name
        if isinstance(vertex, Vertex):
            self[name] = {}
            for state in vertex.possible_vertex_states:
                self[name][state] = None
        else:
            raise TypeError("Vertices must inherit from `Vertex` .")

    def to_hdf(self, hdf, group_name="edges"):
        with hdf.open(group_name) as hdf5_server:
            hdf5_server["TYPE"] = str(type(self))
            for name, edge in self.items():
                hdf5_server[name] = edge

    def from_hdf(self, hdf, group_name):
        return



# coding: utf-8
# Copyright (c) Max-Planck-Institut für Eisenforschung GmbH - Computational Materials Design (CM) Department
# Distributed under the terms of "New BSD License", see the LICENSE file.

from __future__ import print_function


from pyiron_contrib.protocol.generic import CompoundVertex
from pyiron_contrib.protocol.list import SerialList, AutoList, ParallelList
from pyiron_contrib.protocol.primitive.fts_vertices import StringRecenter, StringReflect, PositionsRunningAverage, \
    CentroidsRunningAverageMix, CentroidsSmoothing, CentroidsReparameterization, MilestoningVertex
from pyiron_contrib.protocol.primitive.one_state import InterpolatePositions, RandomVelocity, \
    ExternalHamiltonian, VerletPositionUpdate, VerletVelocityUpdate, \
    Counter, Zeros, WelfordOnline, SphereReflection
from pyiron_contrib.protocol.primitive.two_state import IsGEq, ModIsZero
from pyiron_contrib.protocol.utils import Pointer
import numpy as np
import matplotlib.pyplot as plt
from ase.geometry import find_mic
from scipy.constants import femto as FS

"""
Protocols and vertices for the Finite Temperature String method.
"""

__author__ = "Raynol Dsouza, Liam Huber"
__copyright__ = "Copyright 2019, Max-Planck-Institut für Eisenforschung GmbH " \
                "- Computational Materials Design (CM) Department"
__version__ = "0.0"
__maintainer__ = "Liam Huber"
__email__ = "huber@mpie.de"
__status__ = "development"
__date__ = "23 July, 2019"


class StringRelaxation(CompoundVertex):
    """
    Runs Finite Temperature String relaxation

    Input attributes:
        ref_job_full_path (string): The path to the saved reference job to use for calculating forces and energies.
        structure_initial (Atoms): The initial structure.
        structure_initial (Atoms): The final structure.
        n_images (int): Number of centroids/images.
        n_steps (int): How many steps to run for.
        thermalization_steps (int): How many steps to run for before beginning to update the string.
        sampling_period (int): How often to sample the images and update the centroids. (Default is to synchronize with
            the archiving period; it is highly recommended to set this value to divide the archiving period evenly, so
            that vertices in this branch of the graph get their output archived as normal.)
        temperature (float): Temperature to run at in K.
        reflection_cutoff_distance (float): How far (angstroms) to allow each individual atom to stray from its
            centroid value before reflecting it.
        relax_endpoints (bool): Whether or not to allow string endpoint centroids to evolve. (Default is False.)
        mixing_fraction (float): How much of the images' running average of position to mix into the centroid positions
            each time the mixer is called. (Default is 0.1.)
        nominal_smoothing (float): How much smoothing to apply to the updating centroid positions (endpoints are
            not effected). The actual smoothing is the product of this nominal value, the number of images, and the
            mixing fraction, ala Vanden-Eijnden and Venturoli (2009). (Default is 0.1.)
        temperature_damping_timescale (float): Langevin thermostat timescale in fs. (Default is 100 fs.)
        overheat_fraction (float): The fraction by which to overheat the initial velocities. This can be useful for
            more quickly equilibrating a system whose initial structure is its fully relaxed positions -- in which
            case equipartition of energy tells us that the kinetic energy should be initialized to double the
            desired value. (Default is 2.0, rely on equipartition.)
        time_step (float): MD timestep in fs. (Default is 1 fs.)

    TODO: Wire it so sphere reflection is optional
    """

    def __init__(self, project=None, name=None, job_name=None):
        super(StringRelaxation, self).__init__(project=project, name=name, job_name=job_name)

        # Protocol defaults
        id_ = self.input.default

        id_.relax_endpoints = False
        id_.mixing_fraction = 0.1
        id_.nominal_smoothing = 0.1
        id_.temperature_damping_timescale = 100.
        id_.overheat_fraction = 2.
        id_.time_step = 1.

    def define_vertices(self):
        # Graph components
        g = self.graph
        g.interpolate_positions = InterpolatePositions()
        g.initial_velocities = SerialList(RandomVelocity)
        g.initial_forces = Zeros()
        g.check_steps = IsGEq()
        g.verlet_positions = SerialList(VerletPositionUpdate)
        g.calc_static_images = AutoList(ExternalHamiltonian)
        g.verlet_velocities = SerialList(VerletVelocityUpdate)
        g.reflect_string = SerialList(StringReflect)
        g.reflect_atoms = SerialList(SphereReflection)
        g.check_thermalized = IsGEq()
        g.check_sampling_period = ModIsZero()
        g.running_average = PositionsRunningAverage()
        g.mix = CentroidsRunningAverageMix()
        g.smooth = CentroidsSmoothing()
        g.reparameterize = CentroidsReparameterization()
        g.calc_static_centroids = AutoList(ExternalHamiltonian)
        g.recenter = SerialList(StringRecenter)
        g.clock = Counter()

    def define_execution_flow(self):
        # Execution flow
        g = self.graph
        g.make_pipeline(
            g.interpolate_positions,
            g.initial_velocities,
            g.initial_forces,
            g.check_steps, 'false',
            g.verlet_positions,
            g.reflect_string,  # Comes before atomic reflecting so we can actually trigger a full string reflection!
            g.reflect_atoms,  # Comes after, since even if the string doesn't reflect, on atom might have migrated
            g.calc_static_images,
            g.verlet_velocities,
            g.check_thermalized, 'true',
            g.check_sampling_period, 'true',
            g.running_average,
            g.mix,
            g.smooth,
            g.reparameterize,
            g.calc_static_centroids,
            g.recenter,
            g.clock,
            g.check_steps
        )
        g.make_edge(g.check_thermalized, g.recenter, 'false')
        g.make_edge(g.check_sampling_period, g.recenter, 'false')
        g.starting_vertex = g.interpolate_positions
        g.restarting_vertex = g.check_steps

    def define_information_flow(self):
        # Data flow
        g = self.graph
        gp = Pointer(self.graph)
        ip = Pointer(self.input)

        # interpolate_positions
        g.interpolate_positions.input.structure_initial = ip.structure_initial
        g.interpolate_positions.input.structure_final = ip.structure_final
        g.interpolate_positions.input.n_images = ip.n_images

        # initial_velocities
        g.initial_velocities.input.n_children = ip.n_images
        g.initial_velocities.direct.temperature = ip.temperature
        g.initial_velocities.direct.masses = ip.structure_initial.get_masses
        g.initial_velocities.direct.overheat_fraction = ip.overheat_fraction

        # initial_forces
        g.initial_forces.input.shape = ip.structure_initial.positions.shape

        # check_steps
        g.check_steps.input.target = gp.clock.output.n_counts[-1]
        g.check_steps.input.threshold = ip.n_steps

        # verlet_positions
        g.verlet_positions.input.n_children = ip.n_images
        g.verlet_positions.direct.time_step = ip.time_step
        g.verlet_positions.direct.masses = ip.structure_initial.get_masses
        g.verlet_positions.direct.temperature = ip.temperature
        g.verlet_positions.direct.temperature_damping_timescale = ip.temperature_damping_timescale

        g.verlet_positions.broadcast.default.positions = gp.interpolate_positions.output.interpolated_positions[-1]
        g.verlet_positions.broadcast.default.velocities = gp.initial_velocities.output.velocities[-1]
        g.verlet_positions.direct.default.forces = gp.initial_forces.output.zeros[-1]

        g.verlet_positions.broadcast.positions = gp.recenter.output.positions[-1]
        g.verlet_positions.broadcast.velocities = gp.verlet_velocities.output.velocities[-1]
        g.verlet_positions.broadcast.forces = gp.recenter.output.forces[-1]

        # reflect string
        g.reflect_string.input.n_children = ip.n_images
        g.reflect_string.direct.cell = ip.structure_initial.cell
        g.reflect_string.direct.pbc = ip.structure_initial.pbc

        g.reflect_string.direct.default.all_centroid_positions = \
            gp.interpolate_positions.output.interpolated_positions[-1]
        g.reflect_string.broadcast.default.centroid_positions = \
            gp.interpolate_positions.output.interpolated_positions[-1]
        g.reflect_string.broadcast.default.previous_positions = \
            gp.interpolate_positions.output.interpolated_positions[-1]
        g.reflect_string.broadcast.default.previous_velocities = gp.initial_velocities.output.velocities[-1]

        g.reflect_string.direct.all_centroid_positions = gp.reparameterize.output.centroids_pos_list[-1]
        g.reflect_string.broadcast.centroid_positions = gp.reparameterize.output.centroids_pos_list[-1]
        g.reflect_string.broadcast.previous_positions = gp.recenter.output.positions[-1]
        g.reflect_string.broadcast.previous_velocities = gp.verlet_velocities.output.velocities[-1]

        g.reflect_string.broadcast.positions = gp.verlet_positions.output.positions[-1]
        g.reflect_string.broadcast.velocities = gp.verlet_positions.output.velocities[-1]

        # reflect individual atoms which stray too far
        g.reflect_atoms.input.n_children = ip.n_images
        g.reflect_atoms.broadcast.default.reference_positions = \
            gp.interpolate_positions.output.interpolated_positions[-1]
        g.reflect_atoms.broadcast.default.previous_positions = \
            gp.interpolate_positions.output.interpolated_positions[-1]
        g.reflect_atoms.broadcast.default.previous_velocities = gp.initial_velocities.output.velocities[-1]

        g.reflect_atoms.broadcast.reference_positions = gp.reparameterize.output.centroids_pos_list[-1]
        g.reflect_atoms.broadcast.positions = gp.reflect_string.output.positions[-1]
        g.reflect_atoms.broadcast.velocities = gp.reflect_string.output.velocities[-1]
        g.reflect_atoms.broadcast.previous_positions = gp.recenter.output.positions[-1]
        g.reflect_atoms.broadcast.previous_velocities = gp.verlet_velocities.output.velocities[-1]
        g.reflect_atoms.direct.cutoff_distance = ip.reflection_cutoff_distance
        g.reflect_atoms.direct.cell = ip.structure_initial.cell
        g.reflect_atoms.direct.pbc = ip.structure_initial.pbc

        # calc_static_images
        g.calc_static_images.input.n_children = ip.n_images
        g.calc_static_images.direct.ref_job_full_path = ip.ref_job_full_path
        g.calc_static_images.direct.structure = ip.structure_initial
        g.calc_static_images.broadcast.positions = gp.reflect_atoms.output.positions[-1]

        # verlet_velocities
        g.verlet_velocities.input.n_children = ip.n_images
        g.verlet_velocities.direct.time_step = ip.time_step
        g.verlet_velocities.direct.masses = ip.structure_initial.get_masses
        g.verlet_velocities.direct.temperature = ip.temperature
        g.verlet_velocities.direct.temperature_damping_timescale = ip.temperature_damping_timescale

        g.verlet_velocities.broadcast.velocities = gp.reflect_atoms.output.velocities[-1]
        g.verlet_velocities.broadcast.forces = gp.calc_static_images.output.forces[-1]

        # check_thermalized
        g.check_thermalized.input.target = gp.clock.output.n_counts[-1]
        g.check_thermalized.input.threshold = ip.thermalization_steps

        # check_sampling_period
        g.check_sampling_period.input.target = gp.clock.output.n_counts[-1]
        g.check_sampling_period.input.default.mod = Pointer(self.archive.period)
        g.check_sampling_period.input.mod = ip.sampling_period

        # running_average
        g.running_average.input.default.running_average_list = \
            gp.interpolate_positions.output.interpolated_positions[-1]
        g.running_average.input.running_average_list = gp.running_average.output.running_average_list[-1]
        g.running_average.input.positions_list = gp.reflect_atoms.output.positions[-1]
        g.running_average.input.relax_endpoints = ip.relax_endpoints
        g.running_average.input.cell = ip.structure_initial.cell
        g.running_average.input.pbc = ip.structure_initial.pbc

        # mix
        g.mix.input.default.centroids_pos_list = gp.interpolate_positions.output.interpolated_positions[-1]
        g.mix.input.centroids_pos_list = gp.reparameterize.output.centroids_pos_list[-1]
        g.mix.input.mixing_fraction = ip.mixing_fraction
        g.mix.input.running_average_list = gp.running_average.output.running_average_list[-1]
        g.mix.input.cell = ip.structure_initial.cell
        g.mix.input.pbc = ip.structure_initial.pbc

        # smooth
        g.smooth.input.kappa = ip.nominal_smoothing
        g.smooth.input.dtau = ip.mixing_fraction
        g.smooth.input.all_centroid_positions = gp.mix.output.centroids_pos_list[-1]

        # reparameterize
        g.reparameterize.input.centroids_pos_list = gp.smooth.output.all_centroid_positions[-1]
        g.reparameterize.input.cell = ip.structure_initial.cell
        g.reparameterize.input.pbc = ip.structure_initial.pbc

        # calc_static_centroids
        g.calc_static_centroids.input.n_children = ip.n_images
        g.calc_static_centroids.direct.ref_job_full_path = ip.ref_job_full_path
        g.calc_static_centroids.direct.structure = ip.structure_initial
        g.calc_static_centroids.broadcast.positions = gp.reparameterize.output.centroids_pos_list[-1]

        # recenter
        g.recenter.input.n_children = ip.n_images
        g.recenter.direct.cell = ip.structure_initial.cell
        g.recenter.direct.pbc = ip.structure_initial.pbc

        g.recenter.direct.default.all_centroid_positions = gp.interpolate_positions.output.interpolated_positions[-1]
        g.recenter.broadcast.default.centroid_positions = gp.interpolate_positions.output.interpolated_positions[-1]
        g.recenter.direct.default.centroid_forces = gp.initial_forces.output.zeros[-1]

        g.recenter.direct.all_centroid_positions = gp.reparameterize.output.centroids_pos_list[-1]
        g.recenter.broadcast.centroid_positions = gp.reparameterize.output.centroids_pos_list[-1]
        g.recenter.broadcast.centroid_forces = gp.calc_static_centroids.output.forces[-1]
        g.recenter.broadcast.positions = gp.reflect_atoms.output.positions[-1]
        g.recenter.broadcast.forces = gp.calc_static_images.output.forces[-1]

        self.set_graph_archive_clock(gp.clock.output.n_counts[-1])

    def get_output(self):
        gp = Pointer(self.graph)
        return {
            'energy_pot': ~gp.calc_static_centroids.output.energy_pot[-1],
            'positions': ~gp.reparameterize.output.centroids_pos_list[-1],
            'forces': ~gp.calc_static_centroids.output.forces[-1]
        }

    def plot_potential_energy(self, frame=-1):
        energies = self.graph.calc_static_centroids.archive.output.energy_pot[frame]
        energies -= np.amin(energies)
        plt.plot(energies, marker='o')
        plt.xlabel('Image')
        plt.ylabel('Energy [eV]')
        return energies


class VirtualWork(CompoundVertex):
    """
    Calculates the virtual work to follow a transition path by integrating over the average force of MD images
    constrained to stay within the Voronoi domain of their Voronoi centroid along the path.

    Input attributes:
        structure (Atoms): The initial or final structure whose number and species of atoms, cell, and pbc are those of
            the Voronoi centroids.
        all_centroid_positions (list/numpy.ndarray): The centroid positions along the discretized transition path.
        n_images (int): How many images the path is divided into. TODO: Extract this from `all_centroid_positions`.
        n_steps (int): How many steps to run for.
        thermalization_steps (int): How many steps to run for before beginning to track the average force.
        sampling_period (int): How often to sample the images and update the centroids. (Default is to synchronize with
            the archiving period; it is highly recommended to set this value to divide the archiving period evenly, so
            that vertices in this branch of the graph get their output archived as normal.)
        temperature (float): Temperature to run at in K.
        reflection_cutoff_distance (float): How far (angstroms) to allow each individual atom to stray from its
            centroid value before reflecting it.
        temperature_damping_timescale (float): Langevin thermostat timescale in fs. (Default is 100 fs.)
        overheat_fraction (float): The fraction by which to overheat the initial velocities. This can be useful for
            more quickly equilibrating a system whose initial structure is its fully relaxed positions -- in which
            case equipartition of energy tells us that the kinetic energy should be initialized to double the
            desired value. (Default is 2.0, rely on equipartition.)
        time_step (float): MD timestep in fs. (Default is 1 fs.)

    TODO: Wire it so sphere reflection is optional
    """

    def __init__(self, project=None, name=None, job_name=None):
        super(VirtualWork, self).__init__(project=project, name=name, job_name=job_name)

        # Protocol defaults
        id_ = self.input.default
        id_.time_step = 1.
        id_.overheat_fraction = 2.
        id_.damping_timescale = 100.

        self._displacements = None

    def define_vertices(self):
        # Graph components
        g = self.graph
        g.initial_velocities = SerialList(RandomVelocity)
        g.initial_forces = Zeros()
        g.check_steps = IsGEq()
        g.verlet_positions = SerialList(VerletPositionUpdate)
        g.reflect_string = SerialList(StringReflect)
        g.reflect_atoms = SerialList(SphereReflection)
        g.calc_static = AutoList(ExternalHamiltonian)
        g.verlet_velocities = SerialList(VerletVelocityUpdate)
        g.check_thermalized = IsGEq()
        g.check_sampling_period = ModIsZero()
        g.average_forces = SerialList(WelfordOnline)
        g.average_positions = SerialList(WelfordOnline)
        g.clock = Counter()

    def define_execution_flow(self):
        # Execution flow
        g = self.graph
        g.make_pipeline(
            g.initial_velocities,
            g.initial_forces,
            g.check_steps, 'false',
            g.verlet_positions,
            g.reflect_string,
            g.reflect_atoms,
            g.calc_static,
            g.verlet_velocities,
            g.check_thermalized, 'true',
            g.check_sampling_period, 'true',
            g.average_forces,
            g.average_positions,
            g.clock,
            g.check_steps
        )
        g.make_edge(g.check_thermalized, g.clock, 'false')
        g.make_edge(g.check_sampling_period, g.clock, 'false')
        g.starting_vertex = g.initial_velocities
        g.restarting_vertex = g.check_steps

    def define_information_flow(self):
        # Data flow
        g = self.graph
        gp = Pointer(self.graph)
        ip = Pointer(self.input)

        # initial_velocities
        g.initial_velocities.input.n_children = ip.n_images
        g.initial_velocities.direct.temperature = ip.temperature
        g.initial_velocities.direct.masses = ip.structure.get_masses
        g.initial_velocities.direct.overheat_fraction = ip.overheat_fraction

        # initial_forces
        g.initial_forces.input.shape = ip.structure.positions.shape

        # check_steps
        g.check_steps.input.target = gp.clock.output.n_counts[-1]
        g.check_steps.input.threshold = ip.n_steps

        # verlet_positions
        g.verlet_positions.input.n_children = ip.n_images
        g.verlet_positions.direct.time_step = ip.time_step
        g.verlet_positions.direct.masses = ip.structure.get_masses
        g.verlet_positions.direct.temperature = ip.temperature
        g.verlet_positions.direct.temperature_damping_timescale = ip.temperature_damping_timescale

        g.verlet_positions.broadcast.default.positions = ip.all_centroid_positions
        g.verlet_positions.broadcast.default.velocities = gp.initial_velocities.output.velocities[-1]
        g.verlet_positions.direct.default.forces = gp.initial_forces.output.zeros[-1]

        g.verlet_positions.broadcast.positions = gp.reflect_atoms.output.positions[-1]
        g.verlet_positions.broadcast.velocities = gp.verlet_velocities.output.velocities[-1]
        g.verlet_positions.broadcast.forces = gp.calc_static.output.forces[-1]

        # reflect entire string
        g.reflect_string.input.n_children = ip.n_images
        g.reflect_string.direct.cell = ip.structure.cell
        g.reflect_string.direct.pbc = ip.structure.pbc

        g.reflect_string.broadcast.default.previous_positions = ip.all_centroid_positions
        g.reflect_string.broadcast.default.previous_velocities = gp.initial_velocities.output.velocities[-1]

        g.reflect_string.broadcast.previous_positions = gp.reflect_atoms.output.positions[-1]
        g.reflect_string.broadcast.previous_velocities = gp.reflect_atoms.output.velocities[-1]

        g.reflect_string.broadcast.positions = gp.verlet_positions.output.positions[-1]
        g.reflect_string.broadcast.velocities = gp.verlet_positions.output.velocities[-1]

        g.reflect_string.direct.all_centroid_positions = ip.all_centroid_positions
        g.reflect_string.broadcast.centroid_positions = ip.all_centroid_positions

        # reflect individual atoms which stray too far
        g.reflect_atoms.input.n_children = ip.n_images
        g.reflect_atoms.direct.cutoff_distance = ip.reflection_cutoff_distance
        g.reflect_atoms.direct.cell = ip.structure.cell
        g.reflect_atoms.direct.pbc = ip.structure.pbc

        g.reflect_atoms.broadcast.default.previous_positions = ip.all_centroid_positions
        g.reflect_atoms.broadcast.default.previous_velocities = gp.initial_velocities.output.velocities[-1]

        g.reflect_atoms.broadcast.previous_positions = gp.reflect_atoms.output.positions[-1]
        g.reflect_atoms.broadcast.previous_velocities = gp.reflect_atoms.output.velocities[-1]

        g.reflect_atoms.broadcast.positions = gp.reflect_string.output.positions[-1]
        g.reflect_atoms.broadcast.velocities = gp.reflect_string.output.velocities[-1]

        g.reflect_atoms.broadcast.reference_positions = ip.all_centroid_positions

        # calc_static
        g.calc_static.input.n_children = ip.n_images
        g.calc_static.direct.ref_job_full_path = ip.ref_job_full_path
        g.calc_static.direct.structure = ip.structure

        g.calc_static.broadcast.positions = gp.reflect_atoms.output.positions[-1]

        # verlet_velocities
        g.verlet_velocities.input.n_children = ip.n_images
        g.verlet_velocities.direct.time_step = ip.time_step
        g.verlet_velocities.direct.masses = ip.structure.get_masses
        g.verlet_velocities.direct.temperature = ip.temperature
        g.verlet_velocities.direct.temperature_damping_timescale = ip.temperature_damping_timescale

        g.verlet_velocities.broadcast.velocities = gp.reflect_atoms.output.velocities[-1]
        g.verlet_velocities.broadcast.forces = gp.calc_static.output.forces[-1]

        # check_thermalized
        g.check_thermalized.input.target = gp.clock.output.n_counts[-1]
        g.check_thermalized.input.threshold = ip.thermalization_steps

        # check_sampling_period
        g.check_sampling_period.input.target = gp.clock.output.n_counts[-1]
        g.check_sampling_period.input.default.mod = Pointer(self.archive.period)
        g.check_sampling_period.input.mod = ip.sampling_period

        # average_forces
        g.average_forces.input.n_children = ip.n_images
        g.average_forces.broadcast.sample = gp.calc_static.output.forces[-1]

        # average_positions
        g.average_positions.input.n_children = ip.n_images
        g.average_positions.broadcast.sample = gp.reflect_atoms.output.positions[-1]

        self.set_graph_archive_clock(gp.clock.output.n_counts[-1])

    def get_output(self):
        gp = Pointer(self.graph)
        return {
            'mean_forces': ~gp.average_forces.output.mean[-1],
            'standard_deviations': ~gp.average_forces.output.std[-1],
            'n_samples': ~gp.average_forces.output.n_samples[-1],
        }

    def get_displacements(self, use_average=False):
        if use_average:
            all_positions = np.array(self.graph.average_positions.output.mean[-1])
        else:
            all_positions = np.array(self.input.all_centroid_positions)
        cell = self.input.structure.cell
        pbc = self.input.structure.pbc
        return np.array([
            find_mic(front - back, cell, pbc)[0]
            for front, back
            in zip(all_positions[1:], all_positions[:-1])
        ])

    def get_work_steps(self, frame=-1, use_average=False):
        mean_forces = np.array(self.graph.average_forces.archive.output.mean[frame])
        midpoint_forces = 0.5 * (mean_forces[1:] + mean_forces[:-1])
        displacements = self.get_displacements(use_average=use_average)
        return np.array([np.tensordot(-f, d) for f, d in zip(midpoint_forces, displacements)])

    def get_virtual_work(self, frame=-1, use_average=False):
        work_steps = self.get_work_steps(frame=frame, use_average=use_average)
        mid_pt = int(np.ceil(len(work_steps) / 2.))
        # TODO: account for odd/even number of steps
        return 0.5 * (np.sum(work_steps[:mid_pt]) - np.sum(work_steps[mid_pt:]))


class VirtualWorkFullStep(VirtualWork):
    """
    Reflects after a full velocity Verlet step instead of reflecting at the half-step.
    The advantage is that reflected velocities are *exactly* the negative of the last step's velocities. The
    disadvantage is that the code is a bit uglier because we need to pass around extra force variables in the reflectors
    so that we don't need to repeat the force calculation after reflection...shit, you can't do per-atom reflections
    this way without fucking up the forces...if you reflect anyone you need to recalculate all the forces! (Well, ok,
    you only need to recalulate the reflected atom(s) and their neighbours, but I don't have that sort of optimization
    right now.)
    """

    def define_execution_flow(self):
        # Execution flow
        g = self.graph
        g.make_pipeline(
            g.initial_velocities,
            g.initial_forces,
            g.check_steps, 'false',
            g.verlet_positions,
            g.calc_static,
            g.verlet_velocities,
            g.reflect_string,
            g.reflect_atoms,
            g.check_thermalized, 'true',
            g.check_sampling_period, 'true',
            g.average_forces,
            g.average_positions,
            g.clock,
            g.check_steps
        )
        g.make_edge(g.check_thermalized, g.clock, 'false')
        g.make_edge(g.check_sampling_period, g.clock, 'false')
        g.starting_vertex = g.initial_velocities
        g.restarting_vertex = g.check_steps

    def define_information_flow(self):
        # Data flow
        g = self.graph
        gp = Pointer(self.graph)
        ip = Pointer(self.input)

        # initial_velocities
        g.initial_velocities.input.n_children = ip.n_images
        g.initial_velocities.direct.temperature = ip.temperature
        g.initial_velocities.direct.masses = ip.structure.get_masses
        g.initial_velocities.direct.overheat_fraction = ip.overheat_fraction

        # initial_forces
        g.initial_forces.input.shape = ip.structure.positions.shape

        # check_steps
        g.check_steps.input.target = gp.clock.output.n_counts[-1]
        g.check_steps.input.threshold = ip.n_steps

        # verlet_positions
        g.verlet_positions.input.n_children = ip.n_images
        g.verlet_positions.direct.time_step = ip.time_step
        g.verlet_positions.direct.masses = ip.structure.get_masses
        g.verlet_positions.direct.temperature = ip.temperature
        g.verlet_positions.direct.temperature_damping_timescale = ip.temperature_damping_timescale

        g.verlet_positions.broadcast.default.positions = ip.all_centroid_positions
        g.verlet_positions.broadcast.default.velocities = gp.initial_velocities.output.velocities[-1]
        g.verlet_positions.direct.default.forces = gp.initial_forces.output.zeros[-1]

        g.verlet_positions.broadcast.positions = gp.reflect_atoms.output.positions[-1]
        g.verlet_positions.broadcast.velocities = gp.reflect_atoms.output.velocities[-1]
        g.verlet_positions.broadcast.forces = gp.calc_static.output.forces[-1]

        # calc_static
        g.calc_static.input.n_children = ip.n_images
        g.calc_static.direct.ref_job_full_path = ip.ref_job_full_path
        g.calc_static.direct.structure = ip.structure

        g.calc_static.broadcast.positions = gp.verlet_positions.output.positions[-1]

        # verlet_velocities
        g.verlet_velocities.input.n_children = ip.n_images
        g.verlet_velocities.direct.time_step = ip.time_step
        g.verlet_velocities.direct.masses = ip.structure.get_masses
        g.verlet_velocities.direct.temperature = ip.temperature
        g.verlet_velocities.direct.temperature_damping_timescale = ip.temperature_damping_timescale

        g.verlet_velocities.broadcast.velocities = gp.verlet_positions.output.velocities[-1]
        g.verlet_velocities.broadcast.forces = gp.calc_static.output.forces[-1]

        # reflect entire string
        g.reflect_string.input.n_children = ip.n_images
        g.reflect_string.direct.cell = ip.structure.cell
        g.reflect_string.direct.pbc = ip.structure.pbc

        g.reflect_string.broadcast.default.previous_positions = ip.all_centroid_positions
        g.reflect_string.broadcast.default.previous_velocities = gp.initial_velocities.output.velocities[-1]
        g.reflect_string.direct.default.previous_forces = gp.initial_forces.output.zeros[-1]

        g.reflect_string.broadcast.previous_positions = gp.reflect_atoms.output.positions[-1]
        g.reflect_string.broadcast.previous_velocities = gp.reflect_atoms.output.velocities[-1]
        g.reflect_string.broadcast.previous_forces = gp.reflect_atoms.output.forces[-1]

        g.reflect_string.broadcast.positions = gp.verlet_positions.output.positions[-1]
        g.reflect_string.broadcast.velocities = gp.verlet_velocities.output.velocities[-1]
        g.reflect_string.broadcast.forces = gp.calc_static.output.forces[-1]

        g.reflect_string.direct.all_centroid_positions = ip.all_centroid_positions
        g.reflect_string.broadcast.centroid_positions = ip.all_centroid_positions

        # reflect individual atoms which stray too far
        g.reflect_atoms.input.n_children = ip.n_images
        g.reflect_atoms.direct.cutoff_distance = ip.reflection_cutoff_distance
        g.reflect_atoms.direct.cell = ip.structure.cell
        g.reflect_atoms.direct.pbc = ip.structure.pbc

        g.reflect_atoms.broadcast.default.previous_positions = ip.all_centroid_positions
        g.reflect_atoms.broadcast.default.previous_velocities = gp.initial_velocities.output.velocities[-1]
        g.reflect_atoms.direct.default.previous_forces = gp.initial_forces.output.zeros[-1]

        g.reflect_atoms.broadcast.previous_positions = gp.reflect_atoms.output.positions[-1]
        g.reflect_atoms.broadcast.previous_velocities = gp.reflect_atoms.output.velocities[-1]
        g.reflect_atoms.broadcast.previous_forces = gp.reflect_atoms.output.forces[-1]

        g.reflect_atoms.broadcast.positions = gp.reflect_string.output.positions[-1]
        g.reflect_atoms.broadcast.velocities = gp.reflect_string.output.velocities[-1]
        g.reflect_atoms.broadcast.forces = gp.reflect_string.output.velocities[-1]

        g.reflect_atoms.broadcast.reference_positions = ip.all_centroid_positions

        # check_thermalized
        g.check_thermalized.input.target = gp.clock.output.n_counts[-1]
        g.check_thermalized.input.threshold = ip.thermalization_steps

        # check_sampling_period
        g.check_sampling_period.input.target = gp.clock.output.n_counts[-1]
        g.check_sampling_period.input.default.mod = Pointer(self.archive.period)
        g.check_sampling_period.input.mod = ip.sampling_period

        # average_forces
        g.average_forces.input.n_children = ip.n_images
        g.average_forces.broadcast.sample = gp.calc_static.output.forces[-1]

        # average_positions
        g.average_positions.input.n_children = ip.n_images
        g.average_positions.broadcast.sample = gp.reflect_atoms.output.positions[-1]

        self.set_graph_archive_clock(gp.clock.output.n_counts[-1])


class ConstrainedMD(CompoundVertex):
    """For parallelizing virtual work. Work in progress, not yet functioning."""

    def __init__(self, project=None, name=None, job_name=None):
        super(ConstrainedMD, self).__init__(project=project, name=name, job_name=job_name)

        # Protocol defaults
        id_ = self.input.default
        id_.time_step = 1.
        id_.overheat_fraction = 2.
        id_.damping_timescale = 100.

    def define_vertices(self):
        # Graph components
        g = self.graph
        g.initial_velocities = RandomVelocity()
        g.initial_forces = Zeros()
        g.check_steps = IsGEq()
        g.verlet_positions = VerletPositionUpdate()
        g.reflect_string = StringReflect()
        g.reflect_atoms = SphereReflection()
        g.calc_static = ExternalHamiltonian()
        g.verlet_velocities = VerletVelocityUpdate()
        g.clock = Counter()

    def define_execution_flow(self):
        # Execution flow
        g = self.graph
        g.make_pipeline(
            g.initial_velocities,
            g.initial_forces,
            g.check_steps, 'false',
            g.verlet_positions,
            g.reflect_string,
            g.reflect_atoms,
            g.calc_static,
            g.verlet_velocities,
            g.clock,
            g.check_steps
        )
        g.starting_vertex = g.initial_velocities
        g.restarting_vertex = g.check_steps

    def define_information_flow(self):
        # Data flow
        g = self.graph
        gp = Pointer(self.graph)
        ip = Pointer(self.input)

        # initial_velocities
        g.initial_velocities.input.temperature = ip.temperature
        g.initial_velocities.input.masses = ip.structure.get_masses
        g.initial_velocities.input.overheat_fraction = ip.overheat_fraction

        # initial_forces
        g.initial_forces.input.shape = ip.structure.positions.shape

        # check_steps
        g.check_steps.input.target = gp.clock.output.n_counts[-1]
        g.check_steps.input.threshold = ip.n_steps

        # verlet_positions
        g.verlet_positions.input.time_step = ip.time_step
        g.verlet_positions.input.masses = ip.structure.get_masses
        g.verlet_positions.input.temperature = ip.temperature
        g.verlet_positions.input.temperature_damping_timescale = ip.temperature_damping_timescale

        g.verlet_positions.input.default.positions = ip.centroid_positions
        g.verlet_positions.input.default.velocities = gp.initial_velocities.output.velocities[-1]
        g.verlet_positions.input.default.forces = gp.initial_forces.output.zeros[-1]

        g.verlet_positions.input.positions = gp.reflect_atoms.output.positions[-1]
        g.verlet_positions.input.velocities = gp.verlet_velocities.output.velocities[-1]
        g.verlet_positions.input.forces = gp.calc_static.output.forces[-1]

        # reflect
        g.reflect_string.input.cell = ip.structure.cell
        g.reflect_string.input.pbc = ip.structure.pbc

        g.reflect_string.input.default.previous_positions = ip.centroid_positions
        g.reflect_string.input.default.previous_velocities = gp.initial_velocities.output.velocities[-1]

        g.reflect_string.input.previous_positions = gp.reflect_atoms.output.positions[-1]
        g.reflect_string.input.previous_velocities = gp.reflect_atoms.output.velocities[-1]

        g.reflect_string.input.positions = gp.verlet_positions.output.positions[-1]
        g.reflect_string.input.velocities = gp.verlet_positions.output.velocities[-1]

        g.reflect_string.input.all_centroid_positions = ip.all_centroid_positions
        g.reflect_string.input.centroid_positions = ip.centroid_positions

        # reflect individual atoms which stray too far
        g.reflect_atoms.input.cutoff_distance = ip.reflection_cutoff_distance
        g.reflect_atoms.input.cell = ip.structure.cell
        g.reflect_atoms.input.pbc = ip.structure.pbc

        g.reflect_atoms.input.reference_positions = ip.centroid_positions

        g.reflect_atoms.input.default.previous_positions = ip.centroid_positions
        g.reflect_atoms.input.default.previous_velocities = gp.initial_velocities.output.velocities[-1]

        g.reflect_atoms.input.previous_positions = gp.reflect_atoms.output.positions[-1]
        g.reflect_atoms.input.previous_velocities = gp.reflect_atoms.output.velocities[-1]

        g.reflect_atoms.input.positions = gp.reflect_string.output.positions[-1]
        g.reflect_atoms.input.velocities = gp.reflect_string.output.velocities[-1]

        # calc_static
        g.calc_static.input.ref_job_full_path = ip.ref_job_full_path
        g.calc_static.input.structure = ip.structure

        g.calc_static.input.positions = gp.reflect_atoms.output.positions[-1]

        # verlet_velocities
        g.verlet_velocities.input.time_step = ip.time_step
        g.verlet_velocities.input.masses = ip.structure.get_masses
        g.verlet_velocities.input.temperature = ip.temperature
        g.verlet_velocities.input.temperature_damping_timescale = ip.temperature_damping_timescale

        g.verlet_velocities.input.velocities = gp.reflect_atoms.output.velocities[-1]
        g.verlet_velocities.input.forces = gp.calc_static.output.forces[-1]

        self.set_graph_archive_clock(gp.clock.output.n_counts[-1])

    def get_output(self):
        gp = Pointer(self.graph)
        return {
            'positions': ~gp.reflect_string.output.positions[-1],
            'velocities': ~gp.verlet_velocities.output.velocities[-1],
            'forces': ~gp.calc_static.output.forces[-1]
        }

    def parallel_setup(self):
        super(ConstrainedMD, self).parallel_setup()
        self.graph.calc_static.parallel_setup()


class VirtualWorkParallel(VirtualWork):
    """Work in progress, not yet functioning."""

    def __init__(self, project=None, name=None, job_name=None):
        super(VirtualWorkParallel, self).__init__(project=project, name=name, job_name=job_name)

        # Protocol defaults
        id_ = self.input.default
        id_.time_step = 1.
        id_.overheat_fraction = 2.
        id_.damping_timescale = 100.

        self._displacements = None

    def define_vertices(self):
        # Graph components
        g = self.graph
        g.check_steps = IsGEq()
        g.constrained_evolution = ParallelList(ConstrainedMD)
        g.check_thermalized = IsGEq()
        g.average_forces = SerialList(WelfordOnline)
        g.clock = Counter()

    def define_execution_flow(self):
        # Execution flow
        g = self.graph
        g.make_pipeline(
            g.check_steps, 'false',
            g.constrained_evolution,
            g.check_thermalized, 'true',
            g.average_forces,
            g.clock,
            g.check_steps
        )
        g.make_edge(g.check_thermalized, g.clock, 'false')
        g.starting_vertex = g.check_steps
        g.restarting_vertex = g.check_steps

    def define_information_flow(self):
        # Data flow
        g = self.graph
        gp = Pointer(self.graph)
        ip = Pointer(self.input)

        # check_steps
        g.check_steps.input.target = gp.clock.output.n_counts[-1]
        g.check_steps.input.threshold = ip.n_steps

        # constrained_evolution
        g.constrained_evolution.input.n_children = ip.n_images
        g.constrained_evolution.direct.structure = ip.structure
        g.constrained_evolution.direct.temperature = ip.temperature
        g.constrained_evolution.direct.temperature_damping_timescale = ip.temperature_damping_timescale
        g.constrained_evolution.direct.all_centroid_positions = ip.all_centroid_positions
        g.constrained_evolution.broadcast.centroid_positions = ip.all_centroid_positions
        g.constrained_evolution.direct.reflection_cutoff_distance = ip.reflection_cutoff_distance
        g.constrained_evolution.direct.ref_job_full_path = ip.ref_job_full_path
        g.constrained_evolution.direct.n_steps = ip.sampling_period

        # check_thermalized
        g.check_thermalized.input.target = gp.clock.output.n_counts[-1]
        g.check_thermalized.input.threshold = ip.thermalization_periods

        # average_forces
        g.average_forces.input.n_children = ip.n_images
        g.average_forces.broadcast.sample = gp.constrained_evolution.output.forces[-1]

        self.set_graph_archive_clock(gp.clock.output.n_counts[-1])


class VirtualWorkSerial(VirtualWorkParallel):
    """Work in progress, not yet functioning."""

    def define_vertices(self):
        # Graph components
        g = self.graph
        g.check_steps = IsGEq()
        g.constrained_evolution = SerialList(ConstrainedMD)
        g.check_thermalized = IsGEq()
        g.average_forces = SerialList(WelfordOnline)
        g.clock = Counter()


class Milestoning(CompoundVertex):
    """
    Calculates the jump frequencies of each centroid to the final centroid

    Input attributes:
        structure (Atoms): The initial or final structure to get cell and pbc
        relaxed_centroids_list (list/numpy.ndarray): The centroids on the fully relaxed string
        job_type ('Lammps'/'Vasp'/'Sphinx'): The job type. The listed types are allowed by the code, but may or may not
            be available on your system. (Default is 'Lammps')
        potential (str): The classical potential to use. (Does not exist by default, needs to be set after
            instantiation. Mandatory now, but in the future only needed when the job type is 'Lammps' or another
            classical interpreter that uses potentials.
        n_steps (int): How many steps to run for. (Default is 100.)

        structure (Atoms): The initial or final structure whose number and species of atoms, cell, and pbc are those of
            the Voronoi centroids.
        all_centroid_positions (list/numpy.ndarray): The centroid positions along the discretized transition path.
        n_images (int): How many images the path is divided into. TODO: Extract this from `all_centroid_positions`.
        n_steps (int): How many steps to run for.
        thermalization_steps (int): How many steps to run for before beginning to track transitions.
        temperature (float): Temperature to run at in K.
        temperature_damping_timescale (float): Langevin thermostat timescale in fs. (Default is 100 fs.)
        overheat_fraction (float): The fraction by which to overheat the initial velocities. This can be useful for
            more quickly equilibrating a system whose initial structure is its fully relaxed positions -- in which
            case equipartition of energy tells us that the kinetic energy should be initialized to double the
            desired value. (Default is 2.0, rely on equipartition.)
        time_step (float): MD timestep in fs. (Default is 1 fs.)
    """

    def __init__(self, project=None, name=None, job_name=None):
        super(Milestoning, self).__init__(project=project, name=name, job_name=job_name)

        # Protocol defaults
        id_ = self.input.default
        id_.time_step = 1.
        id_.overheat_fraction = 2.
        id_.damping_timescale = 100.

    def define_vertices(self):
        # Graph components
        g = self.graph
        g.initial_velocities = SerialList(RandomVelocity)
        g.initial_forces = Zeros()
        g.check_steps = IsGEq()
        g.verlet_positions = SerialList(VerletPositionUpdate)
        g.calc_static = AutoList(ExternalHamiltonian)
        g.verlet_velocities = SerialList(VerletVelocityUpdate)
        g.check_thermalized = IsGEq()
        g.milestone = MilestoningVertex()
        g.clock = Counter()

    def define_execution_flow(self):
        # Execution flow
        g = self.graph
        g.make_pipeline(
            g.initial_velocities,
            g.initial_forces,
            g.check_steps, 'false',
            g.verlet_positions,
            g.calc_static,
            g.verlet_velocities,
            g.milestone,
            g.clock,
            g.check_steps
        )
        g.starting_vertex = g.initial_velocities
        g.restarting_vertex = g.check_steps

    def define_information_flow(self):
        # Data flow
        g = self.graph
        gp = Pointer(self.graph)
        ip = Pointer(self.input)

        # initial_velocities
        g.initial_velocities.input.n_children = ip.n_images
        g.initial_velocities.direct.temperature = ip.temperature
        g.initial_velocities.direct.masses = ip.structure.get_masses
        g.initial_velocities.direct.overheat_fraction = ip.overheat_fraction

        # initial_forces
        g.initial_forces.input.shape = ip.structure.positions.shape

        # check_steps
        g.check_steps.input.target = gp.clock.output.n_counts[-1]
        g.check_steps.input.threshold = ip.n_steps

        # verlet_positions
        g.verlet_positions.input.n_children = ip.n_images
        g.verlet_positions.direct.time_step = ip.time_step
        g.verlet_positions.direct.masses = ip.structure.get_masses
        g.verlet_positions.direct.temperature = ip.temperature
        g.verlet_positions.direct.temperature_damping_timescale = ip.temperature_damping_timescale

        g.verlet_positions.broadcast.default.positions = ip.all_centroid_positions
        g.verlet_positions.broadcast.default.velocities = gp.initial_velocities.output.velocities[-1]
        g.verlet_positions.direct.default.forces = gp.initial_forces.output.zeros[-1]

        g.verlet_positions.broadcast.positions = gp.milestone.output.positions[-1]
        g.verlet_positions.broadcast.velocities = gp.milestone.output.velocities[-1]
        g.verlet_positions.broadcast.forces = gp.milestone.output.forces[-1]

        # calc_static
        g.calc_static.input.n_children = ip.n_images
        g.calc_static.direct.ref_job_full_path = ip.ref_job_full_path
        g.calc_static.direct.structure = ip.structure

        g.calc_static.broadcast.positions = gp.verlet_positions.output.positions[-1]

        # verlet_velocities
        g.verlet_velocities.input.n_children = ip.n_images
        g.verlet_velocities.direct.time_step = ip.time_step
        g.verlet_velocities.direct.masses = ip.structure.get_masses
        g.verlet_velocities.direct.temperature = ip.temperature
        g.verlet_velocities.direct.temperature_damping_timescale = ip.temperature_damping_timescale

        g.verlet_velocities.broadcast.velocities = gp.verlet_positions.output.velocities[-1]
        g.verlet_velocities.broadcast.forces = gp.calc_static.output.forces[-1]

        # Milestoning
        g.milestone.input.default.prev_positions_list = ip.all_centroid_positions
        g.milestone.input.default.prev_velocities_list = gp.initial_velocities.output.velocities[-1]
        g.milestone.input.default.prev_forces_list = gp.initial_forces.output.zeros[-1]

        g.milestone.input.prev_positions_list = gp.milestone.output.positions_list[-1]
        g.milestone.input.prev_velocities_list = gp.milestone.output.velocities_list[-1]
        g.milestone.input.prev_forces_list = gp.milestone.output.forces_list[-1]

        g.milestone.input.positions_list = gp.verlet_positions.output.positions[-1]
        g.milestone.input.velocities_list = gp.verlet_velocities.output.velocities[-1]
        g.milestone.input.forces_list = gp.calc_static.output.forces[-1]

        g.milestone.input.all_centroid_positions = ip.all_centroid_positions

        g.milestone.input.thermalization_steps = ip.thermalization_steps
        g.milestone.input.cell = ip.structure.cell
        g.milestone.input.pbc = ip.structure.pbc

        self.set_graph_archive_clock(gp.clock.output.n_counts[-1])

    def get_output(self):
        gp = Pointer(self.graph)
        return {
            # Nothing for now
        }

    def get_transitions(self):
        """
        Generates jump frequencies of each centroid to the final centroid form the reflections matrix, edge reflections
        matrix, and the edge time matrix stored by the milestoning vertex.

        Returns:
            (float): The mean time of first passage from the 0th image to the final image.
            (list): `n_images - 1` mean times of passage from the 0th Voronoi cell to the final cell.
            (list): Respective equilibrium probabilities to find the system in each Voronoi cell.

        TODO: Convert the final frequency to THz?
        """
        time_step = self.input.time_step
        reflections_matrix = self.graph.milestone.output.reflections_matrix[-1]
        edge_reflections_matrix = self.graph.milestone.output.edge_reflections_matrix[-1]
        edge_time_matrix = self.graph.milestone.output.edge_time_matrix[-1]
        # TODO: Add a `frame` input variable and rely on `graph.milestone.archive.output.X[frame]` instead

        n_images = len(reflections_matrix)
        n_edges = int(n_images * (n_images - 1) / 2)
        pis = self._get_pi(reflections_matrix, n_images)

        # for terminology, refer the following paper: https://doi.org/10.1063/1.3129843

        n_ij = np.zeros((n_edges, n_edges))
        r_i = np.zeros(n_edges)
        for img in np.arange(n_images):
            n = pis[img] * edge_reflections_matrix[img]
            r = pis[img] * edge_time_matrix[img]
            n_ij += n
            r_i += r

        # The paper uses shows ways to calculate the mean free passage time. The following section is
        # commented out, as it is redundant. May come in handy for a consistency check?

        # q_ij = []
        # for i in np.arange(n_edges):
        #     if r_i[i] != 0:
        #         q_ij += [n_ij[i] / r_i[i]]
        #     else:
        #         q_ij += [np.zeros(n_edges)]
        #
        # q_ij = np.array(q_ij)  # TODO: pycharm claims this variable is never used again -- remove re-casting or use?

        p_ij = []
        tau_i = []
        n_ji = n_ij.T
        with np.errstate(invalid='ignore'):  # just ignores the divide by zero error
            for i in np.arange(n_edges):
                p_ij += [n_ji[i] / np.sum(n_ji[i])]
                tau_i += [r_i[i] / np.sum(n_ji[i])]
        for i in np.arange(n_edges):
            for j in np.arange(n_edges):
                if np.isnan(p_ij[i][j]):
                    p_ij[i][j] = 0
                    tau_i[i] = 0
        p_ij = np.array(p_ij)
        p_new = np.delete(p_ij, n_edges - 1, 0)
        p_new = np.delete(p_new, n_edges - 1, 1)
        tau_new = np.delete(tau_i, n_edges - 1, 0)

        t_n = np.linalg.lstsq(np.eye(n_edges - 1) - p_new, tau_new, rcond=None)[0] * time_step * FS

        summation = 0
        mean_free_passage_times = [t_n[summation]]  # TODO: Consider longer but more transparent variable names
        for i in np.arange(2, len(t_n)):
            summation += i
            if summation < len(t_n):
                mean_free_passage_times += [t_n[summation]]
        mean_free_passage_times = np.array(mean_free_passage_times)
        jump_frequencies = 1. / mean_free_passage_times

        return jump_frequencies, mean_free_passage_times, pis

    @staticmethod
    def _get_pi(reflections_matrix, n_images):
        dia = np.eye(n_images) * np.sum(reflections_matrix, axis=1)
        pi_mat_a = np.append(reflections_matrix.T - dia, [np.ones(n_images)], axis=0)
        pi_vec_b = np.append(np.zeros(n_images), [1])

        return np.linalg.lstsq(pi_mat_a, pi_vec_b, rcond=None)[0]



"""
Setuptools based setup module
"""
from setuptools import setup, find_packages


setup(
    name='pyiron_contrib',
    version='0.0.1',
    description='Repository for user-generated plugins to the pyiron IDE.',
    long_description='http://pyiron.org',

    url='https://github.com/pyiron/pyiron_contrib',
    author='Max-Planck-Institut für Eisenforschung GmbH - Computational Materials Design (CM) Department',
    author_email='huber@mpie.de',
    license='BSD',

    classifiers=[
        'Development Status :: 4 - Beta',
        'Topic :: Scientific/Engineering :: Physics',
        'License :: OSI Approved :: BSD License',
        'Intended Audience :: Science/Research',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8'
    ],

    keywords='pyiron',
    packages=find_packages(exclude=["*tests*"]),
    install_requires=[
        'ase>=3.19.1',
        'matplotlib>=3.2.1',
        'numpy>=1.19.1',
        'pyiron>=0.3.6',
        'scipy>=1.5.2',
        'seaborn>=0.11.0',
        'scikit-image>=0.17.2',
    ]
)
